/*
 * LineawareIntScanner.cpp
 *
 * Time-stamp: <04/08/30 10:26:04 lars>
 *
 * history: 2004/02/19  1.0  LST created.
 */
#include "LineawareIntScanner.h"

/**
 * Reads an inputstream of integers fast.
 *
 * @author Lars Schmidt-Thieme
 * @version 1.0
 */

LineawareIntScanner::LineawareIntScanner(istream* in) {
  this->in = in;

  bufferSize = DEFAULT_BUFFER_SIZE;

  buf = new char[bufferSize];
  bufLength = (*in).readsome(buf, bufferSize);
  pos = 0;
  while ((((ch = buf[pos++]) < '0') || (ch > '9')) && ch != '\n')
    if (pos >= bufLength) {
      bufLength = (*in).readsome(buf, bufferSize);
      pos = 0;
    }
}

LineawareIntScanner::~LineawareIntScanner() {
  if (buf != NULL)
    delete[] buf;
  /*
  if (buf != 0) {
    cout << "  delete scanners buffer " << endl;
    delete[] buf;
  }
  */
}

bool LineawareIntScanner::hasNextInt() {
  return (ch >= '0' && ch <= '9') || (ch == '\n');
}
int LineawareIntScanner::nextInt() {
  if (ch == '\n') { // detected end of line:
    // skip to next valid char:
    ch = 0;
    if (pos >= bufLength) {
      bufLength = (*in).readsome(buf, bufferSize);
      pos = 0;
    }
    while (bufLength != 0 && ((((ch = buf[pos++]) < '0') || (ch > '9')) && ch != '\n')) {
      if (pos >= bufLength) {
	bufLength = (*in).readsome(buf, bufferSize);
	pos = 0;
      }
    }
    return END_OF_LINE;
  }

  // detected int value:
  int value = ch - '0';
  if (pos >= bufLength) {
    bufLength = (*in).readsome(buf, bufferSize);
    pos = 0;
  }
  while (((ch = buf[pos++]) >= '0') && (ch <= '9')) {
    value = value*10 + (ch - '0');
    if (pos >= bufLength) {
      bufLength = (*in).readsome(buf, bufferSize);
      pos = 0;
    }
  }

  if (ch != '\n') {
    // skip to next valid char:
    if (pos >= bufLength) {
      bufLength = (*in).readsome(buf, bufferSize);
      pos = 0;
    }
    while (bufLength != 0 && ((((ch = buf[pos++]) < '0') || (ch > '9')) && ch != '\n')) {
      if (pos >= bufLength) {
	bufLength = (*in).readsome(buf, bufferSize);
	pos = 0;
      }
    }
  }
  return value;
}
