
#include "Sorting.h"

// #include<iostream>

#define SMALL    7
#define MEDIUM  40
#define TH_INSERT      16  /* insertion sort: threshold */

int min(const int x, const int y) {
  if (x < y)
    return x;
  else 
    return y;
}



int med3(const int x[], const int a, const int b, const int c) {
  int compa = x[a], compb = x[b], compc = x[c];
  if (compa < compb) {  // a < b:
    if (compb < compc)  // a < b < c 
      return b; 
    else if (compa < compc) // a < c < b
      return c;
    else                    // c < a < b
      return a;
  } else {             // b < a:
    if (compc < compb) // c < b < a
      return b;
    else if (compa < compc)   // b < a < c
      return a;               
    else                      // b < c < a
      return c; 
  }
}

/**
 * Sorts the specified sub-array of ints into ascending order.
 */
void quickSort(int x[], const int off, const int len) {
  // Insertion sort on smallest arrays
  int tmp;
  if (len < SMALL) {
    for (int i=off; i<len+off; i++)
      for (int j=i; j>off && x[j-1] > x[j]; j--) {
	
	//if (j-1 < off || j >= off+len)
	//  std::cout << "ERROR: j = " << j << std::endl;

	// swap(x, j, j-1):
	tmp = x[j];
	x[j] = x[j-1];
	x[j-1] = tmp;
      }
    return;
  }

  // Choose a partition element, v
  int m = off + len/2;       // Small arrays, middle element
  if (len > SMALL) {
    int l = off;
    int n = off + len - 1;
    if (len > MEDIUM) {        // Big arrays, pseudomedian of 9
      int s = len/8;
      l = med3(x, l,     l+s, l+2*s);
      m = med3(x, m-s,   m,   m+s);
      n = med3(x, n-2*s, n-s, n);
    }
    m = med3(x, l, m, n); // Mid-size, med of 3
  }
  int v = x[m];

  // Establish Invariant: v* (<v)* (>v)* v*
  int a = off, b = a, c = off + len - 1, d = c;
  while(true) {
    int comparison;
    while (b <= c && (comparison=(x[b] - v)) <=0 ) {
      if (comparison == 0) {
	// swap(x, a++, b):
	tmp = x[a];
	x[a] = x[b];
	x[b] = tmp;
	a++;
      }
      b++;
    }
    while (c >= b && (comparison=(x[c] - v)) >=0 ) {
      if (comparison == 0) {
	// swap(x, c, d--):
	tmp = x[c];
	x[c] = x[d];
	x[d] = tmp;
	d--;
      }
      c--;
    }
    if (b > c)
      break;
    // swap(x, b++, c--):
    tmp = x[b];
    x[b] = x[c];
    x[c] = tmp;
    b++; c--;
  }
	
  // Swap partition elements back to middle
  int s, n = off + len;
  s = min(a-off, b-a  );  
  // vecswap(x, off, b-s, s):
  int i1 = off, i2 = b-s;
  for (int i = 0; i < s; i++, i1++, i2++) {
    tmp = x[i1];
    x[i1] = x[i2];
    x[i2] = tmp;
  }
  s = min(d-c,   n-d-1);  
  // vecswap(x, b,   n-s, s):
  i1 = b; i2 = n-s;
  for (int i = 0; i < s; i++, i1++, i2++) {
    tmp = x[i1];
    x[i1] = x[i2];
    x[i2] = tmp;
  }
  
  // Recursively sort non-partition-elements
  if ((s = b-a) > 1)
    quickSort(x, off, s);
  if ((s = d-c) > 1)
    quickSort(x, n-s, s);
}


// ----------------------------------------------------------------------
// sorting w.r.t. another array:

int med3(const int x[], const int a, const int b, const int c, const int comp[]) {
  int compa = comp[x[a]], compb = comp[x[b]], compc = comp[x[c]];
  if (compa < compb) {  // a < b:
    if (compb < compc)  // a < b < c 
      return b; 
    else if (compa < compc) // a < c < b
      return c;
    else                    // c < a < b
      return a;
  } else {             // b < a:
    if (compc < compb) // c < b < a
      return b;
    else if (compa < compc)   // b < a < c
      return a;               
    else                      // b < c < a
      return c; 
  }
}

/**
 * Sorts the specified sub-array of ints into ascending order
 * w.r.t. a given comparator array.
 */
void quickSort(int x[], const int off, const int len, int comp[]) {
  // Insertion sort on smallest arrays
  int tmp;
  if (len < SMALL) {
    for (int i=off; i<len+off; i++)
      for (int j=i; j>off && comp[x[j-1]] > comp[x[j]]; j--) {
	// swap(x, j, j-1):
	tmp = x[j];
	x[j] = x[j-1];
	x[j-1] = tmp;
      }
    return;
  }

  // Choose a partition element, v
  int m = off + len/2;       // Small arrays, middle element
  if (len > SMALL) {
    int l = off;
    int n = off + len - 1;
    if (len > MEDIUM) {        // Big arrays, pseudomedian of 9
      int s = len/8;
      l = med3(x, l,     l+s, l+2*s, comp);
      m = med3(x, m-s,   m,   m+s, comp);
      n = med3(x, n-2*s, n-s, n, comp);
    }
    m = med3(x, l, m, n, comp); // Mid-size, med of 3
  }
  int v = x[m];

  // Establish Invariant: v* (<v)* (>v)* v*
  int a = off, b = a, c = off + len - 1, d = c;
  while(true) {
    int comparison;
    while (b <= c && (comparison=(comp[x[b]] - comp[v])) <=0 ) {
      if (comparison == 0) {
	// swap(x, a++, b):
	tmp = x[a];
	x[a] = x[b];
	x[b] = tmp;
	a++;
      }
      b++;
    }
    while (c >= b && (comparison=(comp[x[c]] - comp[v])) >=0 ) {
      if (comparison == 0) {
	// swap(x, c, d--):
	tmp = x[c];
	x[c] = x[d];
	x[d] = tmp;
	d--;
      }
      c--;
    }
    if (b > c)
      break;
    // swap(x, b++, c--):
    tmp = x[b];
    x[b] = x[c];
    x[c] = tmp;
    b++; c--;
  }
	
  // Swap partition elements back to middle
  int s, n = off + len;
  s = min(a-off, b-a  );  
  // vecswap(x, off, b-s, s):
  int i1 = off, i2 = b-s;
  for (int i = 0; i < s; i++, i1++, i2++) {
    tmp = x[i1];
    x[i1] = x[i2];
    x[i2] = tmp;
  }
  s = min(d-c,   n-d-1);  
  // vecswap(x, b,   n-s, s):
  i1 = b; i2 = n-s;
  for (int i = 0; i < s; i++, i1++, i2++) {
    tmp = x[i1];
    x[i1] = x[i2];
    x[i2] = tmp;
  }
  
  // Recursively sort non-partition-elements
  if ((s = b-a) > 1)
    quickSort(x, off, s, comp);
  if ((s = d-c) > 1)
    quickSort(x, n-s, s, comp);
}

// ----------------------------------------------------------------------
// heap sort w.r.t. comparator array (adapted from Borgelt's apriori):

void heapsort_sift (int *x, int l, int r, int* comp) {
  int t;                    
  int i;                     

  t = x[l];                   
  i = l +l +1;                  
  do {                          
    if ((i < r)                 
	&&  (comp[x[i]] < comp[x[i+1]]))
      i++;                      
    if (comp[t] >= comp[x[i]])
      break;                
    x[l] = x[i];            
    l = i; i += i +1;          
  } while (i <= r);            
  x[l] = t;
} 

/*--------------------------------------------------------------------*/

void heapsort (int *x, const int len, int* comp) {
  int  l, r;                   
  int t, *v;                

  if (len <= 1) 
    return;        
  l = len >> 1;                
  r = len -1;                  
  while (--l >= 0)             
    heapsort_sift(x, l, r, comp);
  v = x;                      
  while (true) {                  
    // swap(v, 0, r)
    t = v[0]; v[0] = v[r]; v[r] = t;                  
    if (--r <= 0)
      break;       
    heapsort_sift(v, 0, r, comp);
  }                            
}  


// ----------------------------------------------------------------------
// heap sort (adapted from Borgelt's apriori):

void heapsort_sift (int *x, int l, int r) {
  int t;                    
  int i;                     

  t = x[l];                   
  i = l +l +1;                  
  do {                          
    if ((i < r)                 
	&&  (x[i] < x[i+1]))
      i++;                      
    if (t >= x[i])
      break;                
    x[l] = x[i];            
    l = i; i += i +1;          
  } while (i <= r);            
  x[l] = t;
} 

/*--------------------------------------------------------------------*/

void heapsort (int *x, const int len) {
  int  l, r;                   
  int t, *v;                

  if (len <= 1) 
    return;        
  l = len >> 1;                
  r = len -1;                  
  while (--l >= 0)             
    heapsort_sift(x, l, r);
  v = x;                      
  while (true) {                  
    // swap(v, 0, r)
    t = v[0]; v[0] = v[r]; v[r] = t;                  
    if (--r <= 0)
      break;       
    heapsort_sift(v, 0, r);
  }                            
}  


/*--------------------------------------------------------------------*/

static void insertSort_rec (int *x, int len) {                               /* --- recursive part of sort */       \
  int *l, *r;                  
  int y, t;                    
  int  m;                       
   
  do {                          
    l = x; r = l +len -1;       
    if (*l > *r) { t = *l; *l = *r; *r = t; }                       
    y = x[len >> 1];            
    if      (y < *l) 
      y = *l;    
    else if (y > *r) 
      y = *r;    
    while (1) {                 
      while (*++l < y)          
        ;                       
      while (*--r > y)          
        ;                       
      if (l >= r) {             
        if (l <= r) { 
	  l++; r--; 
	} 
	break; 
      }
      t = *l; *l = *r; *r = t;  
    }                           
    m = (int)(x +len -l);       
    len = (int)(r -x +1);       
    if (len > m) {               
      if (m >= TH_INSERT)      
        insertSort_rec(l, m); }           
    else {                     
      if (len >= TH_INSERT)      
        insertSort_rec(x, len);           
      x = l; len = m;          
    }                          
  } while (len >= TH_INSERT);    
}  

/*--------------------------------------------------------------------*/

void insertSort (int *x, int len) {
  int  k;
  int *l, *r;
  int t;     
              
  if (len <= 1)
    return;         
  if (len < TH_INSERT)          
    k = len;                    
  else {                      
    insertSort_rec(x, len);              
    k = TH_INSERT -1;         
  }                           
  for (l = r = x; --k > 0; )
    if (*++r < *l) 
      l = r;     
  r = x;                    
  t = *l; *l = *r; *r = t;    
  while (--len > 0) {           
    t = *++r;                 
    for (l = r; *--l > t; k--)
      l[1] = *l;              
    l[1] = t;                 
  }                           
}

/*--------------------------------------------------------------------*/

/**
 * Search key in sorted array x (with offset offset and length len).
 */
int binarySearch(int key, int* x, int from, int len) {
  int midVal;
  int to = from + len - 1;
  while (from <= to) {
    int mid = (from + to) / 2;
    midVal = x[mid];
    if (midVal < key) 
      from = mid + 1;
    else if (midVal > key) 
      to = mid - 1;
    else 
      return mid; // key found
  }
  return -(from + 1);  // key not found.
}
